#include <X11/Xlib.h>

#include "line.h"

static char *tail_kinsoku[] = { "", "", NULL };
static char *head_kinsoku[] = { "", "", "", "", NULL }; 

#define __USE_GNU
#include <ctype.h>

#define DEFINE_WCTYPE_FUNC(funcname) int wc_##funcname(wchar_t c) { unsigned char str[MB_CUR_MAX]; str[wctomb(str, c)] = '\0'; return str[1] == '\0' ? funcname(*str) : 0; }

DEFINE_WCTYPE_FUNC(isgraph)
DEFINE_WCTYPE_FUNC(isblank)
DEFINE_WCTYPE_FUNC(isalnum)
DEFINE_WCTYPE_FUNC(isascii)

static wchar_t *wcswccat(wchar_t *str, wchar_t c)
{
	wchar_t tmp[MB_CUR_MAX];

	tmp[0] = c;
	tmp[1] = L'\0';
	return wcscat(str, tmp);
}
static int lookup(char **table, wchar_t c)
{
	int i;
	char key[MB_CUR_MAX];

	key[wctomb(key, c)] = '\0';
	for (i = 0; table[i] != NULL; i++) {
		if (strcmp(table[i], key) == 0) {
			return 1;
		}
	}
	return 0;
}

void line_expose_callback(line *self, geometry *area)
{
	t_drawer *d = self->drawer;
	word **p = self->words;
	
	for (; *p != NULL; p++) {
		word *w = *p;
		int chokkaku_pos = w->geo->chokkaku_pos - self->screen_geo->chokkaku_pos;
		int shinkou_pos = w->geo->shinkou_pos - self->screen_geo->shinkou_pos;

		
		if (geometry_overlapped_p(w->geo, area)) {
			t_drawer_clear_area(d,
					    d->ext.width - chokkaku_pos - w->geo->chokkaku_len,
					    0 + shinkou_pos,
					    w->geo->chokkaku_len, w->geo->shinkou_len, False);

			if (w->ctrl_p) 
				continue;

			if (tate_chu_yoko_mode && w->tate_chu_yoko) {
				geometry *geo;
				int x, y;

				geo = word_get_extents(w, self);
				x = geo->chokkaku_pos;
				y = geo->shinkou_pos;
				/* y ɸη׻ϤǤΡ */
				free(geo);
				t_drawer_draw_string2(d,
						      d->ext.width - chokkaku_pos - w->geo->chokkaku_len - x,
						      shinkou_pos + w->geo->shinkou_len - (w->geo->shinkou_len + y),
						      w->str,
						      wcslen(w->str));
			} else {
				geometry *geo;
				geo = t_drawer_get_text_extents(d, w->str, wcslen(w->str));
				t_drawer_draw_string(d,
						     d->ext.width - chokkaku_pos -
						     w->geo->chokkaku_len / 2,
						     shinkou_pos,
						     w->str,
						     wcslen(w->str));
			}

		}
	}
}

typedef enum { Latin, NonLatin, NotAWord } WordType;

typedef int ctype_func(wchar_t);

static int get_word_type(wchar_t *str)
{
	ctype_func *tester[] = { wc_isgraph, wc_isblank, NULL };
	WordType result[] = { Latin, NotAWord, };
	int n;
	

	for (n = 0; tester[n] != NULL; n++) {
		int i;
		int success_p = 1;

		for (i = 0; str[i] != L'\0'; i++) {
			if (!tester[n](str[i]))
				success_p = 0;
		}
		if (success_p) return result[n];
	}

	return NonLatin;
}

static WordType get_next_word_type(word **words, int from, int to)
{
	int n;

	for (n = from; n < to; n++) {
		word *w = words[n];
		WordType type;

		type = get_word_type(w->str);
#if 0
		{
			char *t[] = {"Latin", "NonLatin", "NotAWord"};
			fprintf(stderr, " == %s\n", t[type]);
		}
#endif

		if (type == NotAWord)
			continue;
		else
			return type;
	}
	return NotAWord;
}

static void words_set_attr(word **words, int from, int to)
{
	int n;
	WordType previous = NotAWord;

	for (n = from; n < to; n++) {
		switch (get_next_word_type(words, n, to)) {
		case Latin:
			if (previous == NonLatin || previous == NotAWord)
				if (get_next_word_type(words, n+1, to) == NonLatin)
					words[n]->tate_chu_yoko = 1;
			previous = Latin;
			break;
		case NonLatin:
			previous = NonLatin;
			break;
		case NotAWord:
			previous = NotAWord;
			break;
		}

		switch (words[n]->str[0]) {
		case L'\t':
			words[n]->ctrl_p = 1;
			break;
		}
	}
}

geometry *word_get_extents(word *w, line *l)
{
	geometry *geo;

	if (tate_chu_yoko_mode && w->tate_chu_yoko) {
		geo = t_drawer_get_text_extents2(l->drawer, w->str, wcslen(w->str));
	} else {
		geo = t_drawer_get_text_extents(l->drawer, w->str, wcslen(w->str));
		if (geo->chokkaku_len == 0) {
			geo->chokkaku_len = l->drawer->font_width;
		}
		geo->chokkaku_len *= l->line_height;
	}

	if (w->ctrl_p) switch (w->str[0]) {
	case L'\t':
	{
		int tab_width = l->drawer->font_height / 2 * 8;
		int tab_pos = (l->next_pos->shinkou_pos / tab_width + 1) * tab_width;

		geo->shinkou_len = tab_pos - l->next_pos->shinkou_pos;
		assert(geo->shinkou_len <= tab_width);
	}
	}

	return geo;
}

void word_calc_geometry(word *w, int line_chokkaku_len, line *l)
{
	geometry *geo;

	geo = word_get_extents(w, l);
	geo->shinkou_pos = l->next_pos->shinkou_pos;
	geo->chokkaku_pos = l->next_pos->chokkaku_pos + (line_chokkaku_len - geo->chokkaku_len) / 2;

	if (w->geo)
		free(w->geo);
	w->geo = geo;
}

/* ɽԤ˴ޤޤñ(= words[from < to])Υȥꡣnext_pos û롣 */
static void physical_line_set_geometry(line *self, int from, int to, int line_chokkaku_len)
{
	int i;

	for (i = from; i < to; i++) {
		word *w = self->words[i];
		word_calc_geometry(w, line_chokkaku_len, self);
		self->next_pos->shinkou_pos += w->geo->shinkou_len;
	}
	self->next_pos->chokkaku_pos += line_chokkaku_len;
}

void line_calc_geometry(line *self, int shinkou_max)
{
	int i;
	int first_pos = self->next_pos->chokkaku_pos;
	int line_chokkaku_len = -1;
	int index_linehead;
	int next_shinkou_pos = self->next_pos->shinkou_pos;

	for (index_linehead = i = 0; self->words[i] != NULL; i++) {
		word *w = self->words[i];
		geometry *geo;

		geo = word_get_extents(w, self);
		if (line_chokkaku_len == -1) {
			line_chokkaku_len = geo->chokkaku_len;
		} else if (geo->chokkaku_len > line_chokkaku_len) {
			line_chokkaku_len = geo->chokkaku_len;
		}

		if (next_shinkou_pos + geo->shinkou_len > shinkou_max) {
			physical_line_set_geometry(self, index_linehead, i, line_chokkaku_len);
			index_linehead = i;
			next_shinkou_pos = self->next_pos->shinkou_pos = self->geo->shinkou_pos; /* Ƭ᤹ */
			line_chokkaku_len = -1;
			i--; /* Υ롼פϸߤñ줫 */
			continue;
		}
		next_shinkou_pos += geo->shinkou_len;
		free(geo);
	}
	/*  */
	physical_line_set_geometry(self, index_linehead, i, line_chokkaku_len);
	self->next_pos->shinkou_pos = self->geo->shinkou_pos;

	self->geo->shinkou_len = shinkou_max;
	self->geo->chokkaku_len = self->next_pos->chokkaku_pos - first_pos;
}

static int line_count_words(word **words)
{
	int i;

	for (i = 0; words[i] != NULL; i++)
		;

	return i;
}

void line_add_str(line *self, wchar_t *str)
{
	wchar_t *p = str;
	int n, first;
	wchar_t wstr[100] = L"";

	if (self->words) {
		first = n = line_count_words(self->words);
	} else {
		self->words = calloc(1024, sizeof(word *));
		first = n = 0;
	}


	if (*p == L'\0') {
		self->words[n] = calloc(1, sizeof(word));
		self->words[n]->str = wcsdup(L"");
		return;
	}

#if 1
	while (*p != L'\0') {
		enum { Non, Kinsoku, Latin } catmode = Non;

		self->words[n] = calloc(1, sizeof(word));
	redo:
		if (*p == '\0')
			goto next;

		wcswccat(wstr, *p);

		/* § */
		if (*(p+1) != L'\0') {
			if ((catmode == Non || catmode == Kinsoku)
			    &&
			    locale_eucjp_p
			    &&
			    (lookup(head_kinsoku, *(p+1)) || lookup(tail_kinsoku, *p))) {
				catmode = Kinsoku;
				p++;
				goto redo;
			}
			if ((catmode == Non || catmode == Latin)
			    &&
			    (wc_isalnum(*p) && wc_isalnum(*(p+1)))) {
				catmode = Latin;
				p++;
				goto redo;
			}
		}
		p++;
	next:
		self->words[n]->str = wcsdup(wstr);
		wcscpy(wstr, L"");
		n++;
	}
#else
	while (*p != L'\0') {
		wchar_t tmp[10] = L"";

		self->words[n] = calloc(1, sizeof(word));
		tmp[0] = *p;
		self->words[n]->str = wcsdup(tmp);
		p++;
		n++;
	}
#endif

	words_set_attr(self->words, first, n);

}

line *line_new(wchar_t *str, t_drawer *drawer, geometry *screen_geo)
{
	line *self = calloc(1, sizeof(line));

	self->drawer = drawer;
	line_add_str(self, str);
	self->next_pos = geometry_new();
	self->geo = geometry_new();
	self->line_height = 1.33;
	self->screen_geo = screen_geo;

	return self;
}

void line_print(line *self)
{
	wchar_t *p;
	int i, j = 0;
	word *w;

	while ((w = self->words[j++]) != NULL) {
		p = w->str;
		for (i = 0; i < wcslen(w->str); i++) {
			char c[MB_CUR_MAX];

			c[wctomb(c, p[i])] = '\0';
			fprintf(stderr, "%s", c);
		}
	}
	fprintf(stderr, "\n");
}

void line_set_origin(line *self, geometry *origin) 
{
	self->geo = geometry_copy(origin);
	self->next_pos = geometry_copy(origin);
}
