#include "geometry.h"
#include "tategaki.h"
#include "t_drawer.h"

#ifndef NDEBUG
static void print_om_values(XOM om)
{
	XOMCharSetList cslist;
	int i;

	printf("locale: %s\n", XLocaleOfOM(om));
	XGetOMValues(om,
		     XNRequiredCharSet, &cslist,
		     NULL);

	for (i = 0; i < cslist.charset_count; i++) {
		printf("charset%d: %s\n", i, cslist.charset_list[i]);
	}
}
#endif

int t_drawer_define_cursor(t_drawer *self, unsigned int shape)
{
	Cursor c;
	int rv;

	c = XCreateFontCursor(self->display, shape);
	rv = XDefineCursor(self->display, self->window, c);
	XFlush(self->display);
	return rv;
}

int t_drawer_undefine_cursor(t_drawer *self)
{
	return XUndefineCursor(self->display, self->window);
}
  
int t_drawer_set_font(t_drawer *self, char *fontname, XOrientation orientation)
{
	XFontSetExtents *extents;

	self->oc = XCreateOC(self->om,
		       XNOrientation,  orientation,
		       XNBaseFontName, fontname, 
		       NULL);
	if (!self->oc)
		DIE("failed to load default font\n");

	extents = XExtentsOfFontSet(self->oc);
	/* ȹ⤵ϵդɤ */
	self->font_width = extents->max_logical_extent.height;
	self->font_height = extents->max_logical_extent.width;

	self->oc2 = XCreateOC(self->om,
			      XNOrientation, XOMOrientation_LTR_TTB,
			      XNBaseFontName, fontname,
			      NULL);
	if (!self->oc)
		DIE("failed to load default font\n");

	return 0;
}

geometry *t_drawer_get_text_extents(t_drawer *self, wchar_t *text, int len)
{
	geometry *rv = geometry_new();
	XRectangle overall;

	XwcTextExtents(self->oc, text, len, NULL, &overall);

	rv->shinkou_len = XwcTextEscapement(self->oc, text, len);
	rv->chokkaku_len = overall.height;
	rv->shinkou_pos = overall.y;
	rv->chokkaku_pos = overall.x;


	return rv;
}

geometry *t_drawer_get_text_extents2(t_drawer *self, wchar_t *text, int len)
{
	geometry *rv = geometry_new();
	XRectangle r;

	XwcTextExtents(self->oc2, text, len, NULL, &r);

	rv->shinkou_len = r.height;
	rv->chokkaku_len = r.width;
	rv->shinkou_pos = r.y;
	rv->chokkaku_pos = r.x;

	return rv;
}

t_drawer *t_drawer_new(int global_length, int line_length, char *fontname)
{
	t_drawer *self = calloc(1, sizeof(t_drawer));
	XWMHints hints;
	unsigned long mask;

	self->display = XOpenDisplay("");
	self->screen  = DefaultScreen(self->display);
	self->gc = DefaultGC(self->display, self->screen);
	self->om = XOpenOM(self->display, NULL, NULL, NULL);

#ifndef NDEBUG
	print_om_values(self->om);
#endif

	t_drawer_set_font(self, fontname, XOMOrientation_TTB_RTL);

	self->window  = XCreateSimpleWindow(self->display,
				      RootWindow(self->display, self->screen),
				      0, 0, 640, 480,
				      2,
				      BlackPixel(self->display, self->screen),
				      WhitePixel(self->display, self->screen));
	self->ext.x = 0;
	self->ext.y = 0;
	self->ext.width = 640;
	self->ext.height = 480;

	XStoreName(self->display, self->window, "Tategaki Viewer");

	hints.flags = InputHint;
	hints.input = True;
	XSetWMHints(self->display, self->window, &hints);

	self->protocols = XInternAtom(self->display, "WM_DELETE_WINDOW", True);
	XSetWMProtocols(self->display, self->window, &self->protocols, 1);

	mask = ExposureMask | KeyPressMask | StructureNotifyMask | ButtonPressMask | ButtonReleaseMask | ButtonMotionMask;
	XSelectInput(self->display, self->window, mask);

	XMapWindow(self->display, self->window);

	return self;
}

void t_drawer_draw_string(t_drawer *self, int global_pt, int local_pt, wchar_t *str, int len)
{
	XwcDrawString(self->display, self->window, self->oc, self->gc,
		      global_pt, local_pt,
		      str, len);
}

void t_drawer_draw_string2(t_drawer *self, int global_pt, int local_pt, wchar_t *str, int len)
{
	XwcDrawString(self->display, self->window, self->oc2, self->gc,
		      global_pt, local_pt,
		      str, len);
}

void t_drawer_get_next_event(t_drawer *self, XEvent *event)
{
	while (1) {
		XNextEvent(self->display, event);
		if(XFilterEvent(event, self->window) == True)
			continue;
		return;
	}
}

void t_drawer_destroy(t_drawer *self)
{
	XDestroyWindow(self->display, self->window);
	XCloseDisplay(self->display);
	/* θ塢¾꥽ʤФʤ */
	return;
}

void t_drawer_clear(t_drawer *self)
{
	XClearWindow(self->display, self->window);
}

void t_drawer_copy_area(t_drawer *self, int src_x, int src_y, int width, int height, int dest_x, int dest_y)
{
	XCopyArea(self->display, self->window, self->window, self->gc, src_x, src_y, width, height, dest_x, dest_y);
}


void t_drawer_clear_area(t_drawer *self, int x, int y, int width, int height, Bool exposures)
{
	XClearArea(self->display, self->window, x, y, width, height, exposures);
}
